from .service import Client, Server
from . import utils

def client(locals=None, address=None, init_shell=False):
    """Starts a client.

    If set, *locals* must be a dictionary specifying the context in which
    code will be executed.

    *address* is used to set the address to connect to (e.g. "localhost:1234").

    If *init_shell* is `True`, then the service interpreter will be initialized
    as if started from a shell. This includes updates to *locals*, `sys.argv`,
    and `sys.path`.
    """
    svc = Client(locals, init_shell=init_shell)
    svc.connect(address or utils.DEFAULT_ADDR)

def server(locals=None, address=None, init_shell=False):
    """Starts a server.

    *address* is used to set the address to listen on for connections (e.g.
    "localhost:1234").

    If *init_shell* is `True`, then the service interpreter will be initialized
    as if started from a shell. This includes updates to *locals*, `sys.argv`,
    and `sys.path`.
    """
    svc = Server(locals, init_shell=init_shell)
    svc.serve(address or utils.DEFAULT_ADDR)

def start_client(locals=None, address=None, init_shell=False):
    """Starts a client in a thread.

    Arguments are the same as those for `client`.

    Returns a `ServiceThread` instance.
    """
    svc = Client(locals, init_shell=init_shell)
    svc.start(address or utils.DEFAULT_ADDR)
    return svc

def start_server(locals=None, address=None, init_shell=False):
    """Starts a server in a thread.

    Arguments are the same as those for `server`.

    Returns a `ServiceThread` instance.
    """
    svc = Server(locals, init_shell=init_shell)
    svc.start(address or utils.DEFAULT_ADDR)
    return svc



#! /usr/bin/env python

import os
import sys
import glob
import zipfile

from .lib import logs

BASE_PATH = os.path.join(os.path.dirname(__file__), '..')

def get_path(*names):
    return os.path.abspath(os.path.join(BASE_PATH, *names))

MAIN = """\
from telepythy.__main__ import main
main()
"""

log = logs.get('telepythy.pack')

def pack():
    # don't run if frozen
    if getattr(sys, 'frozen', False):
        return

    src_path = get_path('telepythy/**/*.py')
    dst_path = get_path('telepythy/telepythy_service.pyz')

    with zipfile.ZipFile(dst_path, 'w', zipfile.ZIP_DEFLATED,
            allowZip64=False) as zip:
        for path in glob.iglob(src_path, recursive=True):
            if 'gui' in path: continue
            dst = os.path.relpath(path)
            log.debug('+ %s', dst)
            zip.write(path, dst)
        zip.writestr('__main__.py', MAIN)

    log.debug('packed lib stored to: %s', dst_path)

    return dst_path

if __name__ == '__main__':
    try:
        logs.init(2, format='%(message)s')
        pack()
    except KeyboardInterrupt:
        pass



import sys
from logging import *

iswindows = sys.platform == 'win32'

get = getLogger
log = get(__name__)

def init(verbose=0, mode=None, format=None, color=False, set_excepthook=False):
    """Initializes simple logging defaults."""

    level = {
        0: WARNING,
        1: INFO,
        2: DEBUG,
        }[verbose]

    root_log = get()
    # don't increase the level if it's already been set
    root_log.setLevel(min(level, root_log.level))

    tele_log = get('telepythy')
    # do not propagate logs to parent
    tele_log.propagate = False
    tele_log.setLevel(level)

    fmt = format or '%(levelname).1s %(asctime)s [%(mode)s:%(name)s] %(message)s'

    if sys.stderr is None:
        if not iswindows:
            # TODO: maybe try syslog on linux?
            return
        handler = DbgViewHandler()
        formatter = Formatter(fmt)
    else:
        handler = StreamHandler()
        formatter = Formatter(fmt)

        if color:
            # try to enable color formatting
            try:
                import colorlog
            except ImportError:
                # fail silently. no color
                pass
            else:
                formatter = colorlog.ColoredFormatter('%(log_color)s' + fmt)

    handler.setFormatter(formatter)
    handler.addFilter(ModeFilter(mode))

    tele_log.addHandler(handler)

    if set_excepthook:
        sys.excepthook = handle_exception

def handle_exception(etype, evalue, etb):
    if not issubclass(etype, Exception):
        # default handling for BaseExceptions
        sys.__excepthook__(etype, evalue, etb)
        return
    log.error('unhandled exception', exc_info=(etype, evalue, etb))

class ModeFilter(Filter):
    def __init__(self, mode):
        super(ModeFilter, self).__init__()
        self._mode = mode

    def filter(self, record):
        record.mode = self._mode
        return True

if iswindows:
    import ctypes as ct
    from ctypes import wintypes as wt

    OutputDebugString = ct.windll.kernel32.OutputDebugStringW
    OutputDebugString.argtypes = [wt.LPCWSTR]
    OutputDebugString.restype = None

    class DbgViewHandler(Handler):
        def emit(self, record):
            OutputDebugString(self.format(record))



from __future__ import print_function

import sys
import threading
import traceback
try:
    import queue
except ImportError:
    import Queue as queue

from . import logs
from . import utils
from . import sockio
from . import interpreter
from . import event_handlers

Q_TIMEOUT = 0.1

log = logs.get(__name__)

class Service(object):
    """Base class for client/server services."""
    def __init__(self, locals=None, filename=None, init_shell=False):
        self._timeout = Q_TIMEOUT

        self._thread = None
        self._stop = threading.Event()
        self._shutdown = threading.Event()

        self._events = queue.Queue()
        self._code_queue = queue.Queue()

        self._event_handlers = event_handlers.default_handlers()

        self._is_evaluating = False

        if init_shell:
            # set up a shell environment
            locals = locals or {}
            locals.setdefault('__name__', '__main__')

            sys.argv = ['']
            if not sys.path or sys.path[0] != '':
                sys.path.insert(0, '')

        self._inter = interpreter.Interpreter(locals, filename,
            lambda text: self.add_event('stdout', text=text),
            lambda text: self.add_event('stderr', text=text),
            )

    ## threading ##

    def join(self, timeout=None):
        if self._thread is None:
            raise ServiceError('thread is not running')
        self._thread.join(timeout)
        if self._thread.is_alive():
            raise Timeout()

    ## execution ##

    def run(self):
        shutdown = self._shutdown
        shutdown.clear()

        q = self._code_queue
        timeout = self._timeout

        try:
            while not shutdown.is_set():
                for handler in self._event_handlers:
                    handler()

                try:
                    data = q.get(timeout=timeout)
                except queue.Empty:
                    continue

                self.evaluate(**data)
        finally:
            self._thread = None

    def stop(self):
        self._stop.set()
        self._shutdown.set()

    ## interpreter ##

    @property
    def locals(self):
        return self._inter.locals

    def evaluate(self, source, notify=True):
        self._is_evaluating = True
        with self._inter.hooked():
            try:
                self._inter.evaluate(source)
            except (Exception, KeyboardInterrupt):
                traceback.print_exc()
            finally:
                self._is_evaluating = False
                if notify:
                    self.add_event('done')

    def interrupt(self):
        if self._is_evaluating:
            utils.interrupt()

    def complete(self, prefix):
        matches = self._inter.complete(prefix)
        self.add_event('completion', matches=matches)

    def reset(self):
        self._inter.reset()

    ## handlers ##

    def add_event(self, name, **data):
        if name == 'stdout':
            log.debug('out: %r', data['text'][:100])
        elif name == 'stderr':
            log.debug('err: %r', data['text'][:100])
        else:
            log.debug('evt: %s%s', name, (data or '') and ': ' + repr(data))

        event = {'evt': name, 'data': data}
        self._events.put(event)

    def register_event_handler(self, handler):
        self._event_handlers.append(handler)

    def _handle(self, sock):
        self._stop.clear()

        self.add_event('start', version=sys.version)

        t_evt = utils.start_thread(self._handle_events, sock)
        t_cmd = utils.start_thread(self._handle_commands, sock)

        t_evt.join()
        t_cmd.join()

    def _handle_events(self, sock):
        stop = self._stop
        events = self._events
        timeout = self._timeout

        try:
            while not stop.is_set():
                try:
                    event = events.get(timeout=timeout)
                except queue.Empty:
                    sock.sendmsg(None)
                else:
                    sock.sendmsg(event)
        except sockio.error as e:
            log.error('handle_events error: %s', repr(e))
            stop.set()

    def _handle_commands(self, sock):
        stop = self._stop

        try:
            while not stop.is_set():
                try:
                    msg = sock.recvmsg()
                except sockio.timeout:
                    continue

                cmd = msg['cmd']
                data = msg.get('data')
                log.debug('cmd: %s%s', cmd, ': ' + repr(data) if data else '')

                if cmd == 'evaluate':
                    if self._is_evaluating:
                        self._inter.recv_input(data['source'] + '\n')
                    else:
                        self._code_queue.put(data)
                elif cmd == 'interrupt':
                    self.interrupt()
                elif cmd == 'complete':
                    self.complete(data)
                else:
                    log.error('unknown command: %s', cmd)

        except sockio.error as e:
            log.error('handle_commands error: %s', repr(e))
            stop.set()

class Client(Service):
    def start(self, addr):
        if self._thread is not None:
            raise ServiceError('thread already running')
        self._thread = utils.start_thread(self.connect, addr)

    def connect(self, address):
        address = utils.parse_address(address)
        client, _addr = sockio.start_client(address, self._handle)
        try:
            self.run()
        finally:
            client.stop()
            client.join()

class Server(Service):
    def start(self, addr):
        if self._thread is not None:
            raise ServiceError('thread already running')
        self._thread = utils.start_thread(self.serve, addr)

    def serve(self, address):
        address = utils.parse_address(address)
        server, _addr = sockio.start_server(address, self._handle)
        try:
            self.run()
        finally:
            server.stop()
            server.join()

class ServiceError(Exception):
    """Raised for Service errors."""

class Timeout(ServiceError):
    """Raised for join timeouts."""



import io
import json
import time
import errno
import socket
import struct
import threading

from . import logs
from . import utils

TIMEOUT = 0.1
BACKLOG = socket.SOMAXCONN
CHUNK_SIZE = io.DEFAULT_BUFFER_SIZE

error = socket.error
timeout = socket.timeout

log = logs.get(__name__)

def start_client(address, handler, stop=None, retry_limit=-1, retry_interval=1):
    stop = stop or threading.Event()

    t = utils.start_thread(client_loop,
        address, handler, stop, retry_limit, retry_interval)
    return (StoppableThread(t, stop), address)

def client_loop(address, handler, stop, retry_limit, retry_interval):
    count = 0
    timeout = TIMEOUT

    while not stop.is_set():
        try:
            with connect(address, timeout) as sock:
                handler(sock)
        except socket.error as e:
            log.error('connection error: %s', e)

        if stop.is_set():
            break
        count += 1
        if retry_limit != -1 and count > retry_limit:
            log.warning('retry limit reached (attempt #%s)', count)
            break
        time.sleep(retry_interval)
        log.warning('retrying connection (attempt #%s)', count)

def start_server(address, handler, stop=None, backlog=None):
    stop = stop or threading.Event()

    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    sock.bind(address)
    sock.listen(backlog or BACKLOG)

    host, port = sock.getsockname()
    log.info('listening: %s:%s', host, port)

    t = utils.start_thread(server_loop, sock, handler, stop)
    return (StoppableThread(t, stop), (host, port))

def server_loop(server_sock, handler, stop):
    timeout = TIMEOUT
    server_sock.settimeout(timeout)

    while not stop.is_set():
        try:
            s, addr = server_sock.accept()
        except socket.timeout:
            continue

        log.info('connected: %s:%s', *addr)
        with SockIO(s) as sock:
            sock.settimeout(timeout)
            handler(sock)

def connect(address, timeout=None):
    log.debug('connecting: %s:%s', *address)
    sock = socket.create_connection(address, timeout)
    log.info('connected: %s:%s', *address)
    return SockIO(sock)

class SockIO(object):
    def __init__(self, sock, chunk_size=None):
        self._sock = sock
        self._chunk_size = chunk_size or CHUNK_SIZE

    def sendmsg(self, msg):
        data = json.dumps(msg).encode('utf8')
        self.send(data)

    def recvmsg(self):
        data = self.recv()
        return json.loads(data.decode('utf8'))

    def send(self, data):
        data_len = len(data)
        size = struct.pack('>I', data_len)
        self._sock.sendall(size)
        self._sock.sendall(data)

    def recv(self):
        return b''.join(self.recviter())

    def recviter(self):
        buf = b''.join(self.recvsize(4))
        data_len = struct.unpack('>I', buf)[0]
        for chunk in self.recvsize(data_len):
            yield chunk

    def recvsize(self, size):
        sock = self._sock

        pos = 0
        chunk_size = min(size, self._chunk_size)
        while pos < size:
            chunk = sock.recv(min(size-pos, chunk_size))
            if not chunk:
                raise ReceiveInterrupted()
            pos += len(chunk)
            yield chunk

    def settimeout(self, t):
        self._sock.settimeout(t)

    def close(self):
        try:
            self._sock.shutdown(socket.SHUT_RDWR)
        except (OSError, socket.error) as e:
            # ignore if not connected
            if e.errno not in (errno.ENOTCONN,):
                raise
        self._sock.close()

    def __enter__(self):
        return self

    def __exit__(self, etype, evalue, etb):
        self.close()

class StoppableThread(object):
    def __init__(self, thread, stop):
        self._thread = thread
        self._stop = stop

    def stop(self):
        self._stop.set()

    def join(self):
        self._thread.join()

class SockIOError(Exception):
    pass

class ReceiveInterrupted(SockIOError, error):
    pass



import argparse

from . import logs
from . import utils
from . import client, server

log = logs.get(__name__)

def main():
    parser = argparse.ArgumentParser('telepythy',
        description='This service can run as either an client (-c) or '
            'server (-s).\nThe default is to run as a server.')

    group = parser.add_mutually_exclusive_group()
    group.add_argument('-s', '--serve', nargs='?', default=False,
        help='[default] <interface>:<port> to bind to. '
             'set port to 0 for a random port (default: {})'.format(
            utils.DEFAULT_ADDR))
    group.add_argument('-c', '--connect', nargs='?', default=False,
        help='<host>:<port> to connect to (default: {})'.format(
            utils.DEFAULT_ADDR))

    parser.add_argument('-v', '--verbose', action='count',
        default=0, help='enable verbose output (-vv for more)')

    args = parser.parse_args()

    logs.init(args.verbose, mode='svc')

    # enable reception of ctrl+c events as part of a new console group (win)
    utils.set_console_ctrl_handler()

    # serve unless connect is set
    if args.connect is not False:
        client(address=args.connect, init_shell=True)
    else:
        server(address=args.serve, init_shell=True)

if __name__ == '__main__':
    try:
        main()
    except KeyboardInterrupt:
        pass



import sys

DELAY = 0.1 # seconds

def default_handlers():
    return [handle_pyside6]

def handle_pyside6():
    try:
        QtCore = sys.modules['PySide6'].QtCore
    except (KeyError, AttributeError):
        return

    app = QtCore.QCoreApplication.instance()
    if app is None:
        return

    app.processEvents(QtCore.QEventLoop.AllEvents, int(DELAY * 1000))
    timer = QtCore.QTimer()
    event_loop = QtCore.QEventLoop()
    timer.timeout.connect(event_loop.quit)
    timer.start(int(DELAY * 1000))
    event_loop.exec_()
    timer.stop()



import os
import sys
import signal
import threading

from . import logs

BASE_PATH = os.path.abspath(os.path.join(os.path.dirname(__file__), '..', '..'))

DEFAULT_COMMAND = 'python'
DEFAULT_HOST = 'localhost'
DEFAULT_PORT = 7373
DEFAULT_ADDR = '{}:{}'.format(DEFAULT_HOST, DEFAULT_PORT)

IS_WINDOWS = sys.platform == 'win32'

log = logs.get(__name__)

def get_path(*names):
    return os.path.join(BASE_PATH, *names)

def parse_address(address):
    s = address.split(':', 1)
    host = s[0].strip() or DEFAULT_HOST
    if len(s) == 1 or not s[1]:
        return (host, DEFAULT_PORT)
    return (host, int(s[1]))

def start_thread(func, *args, **kwargs):
    def thread(func, *args, **kwargs):
        ident = threading.current_thread().ident
        func_name = getattr(func, '__qualname__', func.__name__)

        log.debug('thread started [%s] (%s)', ident, func_name)
        try:
            return func(*args, **kwargs)
        except:
            log.exception('unexpected thread error [%s] (%s)', ident, func_name)
        finally:
            log.debug('thread stopped [%s] (%s)', ident, func_name)

    t = threading.Thread(target=thread, args=(func,) + args, kwargs=kwargs)
    t.daemon = True
    t.start()
    return t

def interrupt(pid=None):
    pid = os.getpid() if pid is None else pid
    log.debug('interrupting process: %s', pid)
    os.kill(pid, signal.CTRL_C_EVENT if IS_WINDOWS else signal.SIGINT)

if IS_WINDOWS:
    # Handling Ctrl+C cleanly on Windows for child processes is tricky.
    # Dreampie uses a commonly recommended technique that calls
    # GenerateConsoleCtrlEvent, sending SIGINT to every process in the process
    # group, including the parent. It works around this by having the parent
    # sleep in order to catch KeyboardInterrupt and ignore it. Not too clean
    # but it does work.
    #
    # Another approach is to use CREATE_NEW_PROCESS_GROUP for the child,
    # ensuring that the parent will not get the signal. Unfortunately, it
    # seemed like GenerateConsoleCtrlEvent would not work if you did that,
    # and people came up with a lot of hacky workarounds for this problem.
    #
    # Then I stumbled on this buried StackOverflow comment:
    # https://stackoverflow.com/questions/7085604/sending-c-to-python-subprocess-objects-on-windows#comment99085177_7980368
    #
    # The child simply needs to enable Ctrl+C handling with
    # SetConsoleCtrlHandler. So now all is right in the universe (except for
    # everything else that isn't).

    import ctypes as ct
    from ctypes import wintypes as wt

    HandlerRoutine = ct.WINFUNCTYPE(wt.BOOL, wt.DWORD)

    SetConsoleCtrlHandler = ct.windll.kernel32.SetConsoleCtrlHandler
    SetConsoleCtrlHandler.argtypes = (HandlerRoutine, wt.BOOL)
    SetConsoleCtrlHandler.restype = wt.BOOL

    def set_console_ctrl_handler():
        if not SetConsoleCtrlHandler(HandlerRoutine(0), False):
            log.warning('failed to set console ctrl handler')

else:
    def set_console_ctrl_handler():
        pass



import os
from setuptools import setup, find_packages

os.chdir(os.path.dirname(__file__))

import telepythy

with open('README.md') as f:
    readme = f.read()

setup(
    name='telepythy-service',
    version=telepythy.__version__,
    url='https://github.com/dhagrow/telepythy',
    author='Miguel Turner',
    author_email='cymrow@gmail.com',
    long_description=readme,
    long_description_content_type='text/markdown',
    packages=find_packages(),
    entry_points={
        'console_scripts': [
            'telepythy-service=telepythy.lib.__main__:main'
            ]
        },
    )



import ast
import sys
import pprint
import keyword
import itertools
import linecache
import threading
import contextlib
import collections

try:
    import builtins
except ImportError:
    import __builtin__ as builtins
_builtins = builtins.__dict__

from . import logs

log = logs.get(__name__)

try:
    range = xrange
except NameError:
    pass

class Interpreter(object):
    def __init__(self, locals=None, filename=None,
            stdout_callback=None, stderr_callback=None):

        self.locals = {}
        self._init_locals = locals or {}

        self.filename = filename or 'telepythy'

        self._run_lock = threading.Lock()
        self._result_count = 0
        self._result_limit = 30

        self._block_counter = itertools.count()

        self._last_result = None
        self._stdout_callback = stdout_callback
        self._stderr_callback = stderr_callback

        self.reset()

    def reset(self):
        self.locals.clear()
        exec('', self.locals)
        self.locals.update(self._init_locals)
        self._result_count = 0

    ## commands ##

    def evaluate(self, source):
        block = next(self._block_counter)
        fname = '<{}:{}>'.format(self.filename, block)

        # technique from: https://stackoverflow.com/questions/47183305/file-string-traceback-with-line-preview
        # (size, mtime, lines, fullname)
        linecache.cache[fname] = (
            len(source), None, source.splitlines(True), fname)

        with self._run_lock:
            mod = compile(source, fname, 'exec', ast.PyCF_ONLY_AST)
            inter = ast.Interactive(mod.body)
            codeob = compile(inter, fname, 'single')
            exec(codeob, self.locals)

            self._store_result()

    def complete(self, prefix):
        matches = []
        # log.error('prefix: %s', prefix)

        if '.' not in prefix:
            # return global matches
            for kw in keyword.kwlist:
                matches.append(kw)

            for name in self.locals.keys():
                matches.append(name)

            for name in _builtins.keys():
                matches.append(name)
        else:
            ctx = prefix.rsplit('.', 1)[0]
            try:
                obj = eval(ctx, self.locals)
            except Exception:
                # log.error('ctx failed to resolve: %r', ctx)
                pass
            else:
                # log.error('obj: %s', obj)

                for name in dir(obj):
                    matches.append(name)

        return matches

    def complete_x(self, prefix):
        matches = []

        if '.' not in prefix:
            prefix_0 = prefix and prefix[0]

            for kw in keyword.kwlist:
                if kw.startswith(prefix_0):
                    matches.append(kw)

            for name in self.locals.keys():
                if name.startswith(prefix_0):
                    matches.append(name)

            for name in _builtins.keys():
                if name.startswith(prefix_0):
                    matches.append(name)

        else:
            idents = prefix.split('.')
            parents, prefix = idents[:-1], idents[-1]

            obj = self.locals.get(parents[0])
            if not obj:
                obj = _builtins.get(parents[0])
                if not obj:
                    return []

            for ident in parents[1:]:
                try:
                    obj = getattr(obj, ident)
                except AttributeError:
                    return []

            prefix_0 = prefix and prefix[0]

            for name in dir(obj):
                if name.startswith(prefix_0):
                    matches.append(name)

        return sorted(matches, key=match_sort_key)

    ## io control ##

    @contextlib.contextmanager
    def hooked(self):
        self.hook()
        try:
            yield
        finally:
            self.unhook()

    def hook(self):
        sys.stdin = InputIO()
        sys.stdout = OutputIO(self._stdout_callback)
        sys.stderr = OutputIO(self._stderr_callback)

        sys.displayhook = self.displayhook

    def unhook(self):
        sys.stdin = sys.__stdin__
        sys.stdout = sys.__stdout__
        sys.stderr = sys.__stderr__

        sys.displayhook = sys.__displayhook__

    def recv_input(self, text):
        sys.stdin.write(text)

    def displayhook(self, value):
        self._last_result = value

    def _store_result(self):
        if self._last_result is None:
            return

        # pop
        result, self._last_result = self._last_result, None

        self.locals['_'] = result
        self.locals['_{}'.format(self._result_count)] = result

        print('{}: {}'.format(self._result_count, pprint.pformat(result)))

        self._result_count += 1

        # remove old results
        for i in range(max(0, self._result_count-self._result_limit)):
            self.locals.pop('_{}'.format(i), None)

class InputIO:
    def __init__(self):
        self._buffer = collections.deque()
        self._ready = threading.Event()

    def read(self, size):
        buf = self._buffer
        ready = self._ready

        while True:
            if len(buf) < size:
                ready.wait()
                ready.clear()
                continue
            return ''.join(buf.popleft() for _ in range(size))

    def readline(self):
        buf = []
        while True:
            c = self.read(1)
            if c == '\n':
                return ''.join(buf)
            else:
                buf.append(c)

    def write(self, data):
        self._buffer.extend(data)
        self._ready.set()

class OutputIO(object):
    def __init__(self, callback, mirror=None):
        self._callback = callback
        self._mirror = mirror

    def isatty(self):
        # XXX: disabled until support for ANSI codes can be implemented
        return False

    def write(self, s):
        if self._mirror is not None:
            self._mirror.write(s)
        if isinstance(s, bytes):
            s = s.decode('utf8')
        self._callback(s)

    def flush(self):
        if self._mirror is not None:
            self._mirror.flush()

def match_sort_key(match):
    return (match.startswith('_'), match)
